from satella.coding.algos import merge_dicts
from satella.exceptions import ConfigurationError
from .base import BaseSource

__all__ = [
    'AlternativeSource', 'OptionalSource', 'MergingSource'
]


class AlternativeSource(BaseSource):
    """
    If first source of configuration fails with ConfigurationError, use the next one instead, ad
    nauseam.
    """
    __slots__ = ('sources',)

    def __init__(self, *sources: BaseSource):
        super().__init__()
        self.sources = sources  # type: tp.List[BaseSource]

    def __repr__(self) -> str:
        return 'AlternativeSource(%s)' % (repr(self.sources),)

    def provide(self) -> dict:
        """
        :raises ConfigurationError: when backup fails too
        """
        for source in self.sources:
            try:
                s = source.provide()
                assert isinstance(s, dict), 'provide() returned a non-dict'
                return s
            except ConfigurationError:
                pass
        else:
            raise ConfigurationError('all sources failed!')


class OptionalSource(AlternativeSource):
    """
     This will substitute for empty dict if underlying config would fail.

     Apply this to your sources if you expect that they will fail.

     Use as

     >>> OptionalSource(SomeOtherSource1)
     """

    def __init__(self, source: BaseSource):
        super().__init__(source, BaseSource())

    def __repr__(self) -> str:
        return 'OptionalSource(%s)' % (repr(self.sources[0], ))


class MergingSource(BaseSource):
    """
    Source that merges configuration from a bunch of sources. The configuration has to be a
    dictionary!!

    :param sources: Sources to examine. Source later in queue will override earlier's entries, so
        take care.
    :param on_fail: how to behave when a source fails
    :param fail_if_no_sources_are_correct: even if on_fail == MergingSource.SILENT,
        if all sources fail, this will fail as well. Of course this makes sense only if on_fail ==
        MergingSource.SILENT
    """

    RAISE = 0  # Raise ConfigurationError if one of sources fails
    SILENT = 1  # Silently continue loading from next files if one fails
    __slots__ = ('sources', 'on_fail', 'fail_if_no_sources_are_correct')

    def __init__(self, *sources: BaseSource, on_fail: int = RAISE,
                 fail_if_no_sources_are_correct: bool = True):
        super().__init__()
        self.sources = sources
        self.on_fail = on_fail
        self.fail_if_no_sources_are_correct = fail_if_no_sources_are_correct

    def provide(self) -> dict:
        cfg = {}
        correct_sources = 0

        for source in self.sources:
            try:
                p = source.provide()
                correct_sources += 1
            except ConfigurationError as e:
                if self.on_fail == MergingSource.RAISE:
                    raise e
                elif self.on_fail == MergingSource.SILENT:
                    p = {}
                else:
                    raise ConfigurationError('Invalid on_fail parameter %s' % (self.on_fail,))
            assert isinstance(p, dict), 'what was provided by the config was not a dict'
            cfg = merge_dicts(cfg, p)
            assert isinstance(cfg, dict), 'what merge_dicts returned wasn''t a dict'

        if correct_sources == 0 and self.sources and self.fail_if_no_sources_are_correct:
            raise ConfigurationError('No source was able to load the configuration')

        return cfg

    def __repr__(self) -> str:
        return '<MergingSource %s>' % (repr(self.sources),)



import re
import typing as tp

from satella.exceptions import ConfigurationValidationError
from .base import Descriptor, ConfigDictValue
from .registry import register_custom_descriptor


@staticmethod
def _make_boolean(v: tp.Any) -> bool:
    if isinstance(v, str):
        if v.upper() == 'TRUE':
            return True
        elif v.upper() == 'FALSE':
            return False
        else:
            raise ConfigurationValidationError('Unknown value of "%s" posing to be a bool'
                                               % (v,))
    else:
        return bool(v)


@register_custom_descriptor('bool')
class Boolean(Descriptor):
    """
    This value must be a boolean, or be converted to one
    """
    BASIC_MAKER = _make_boolean


@register_custom_descriptor('int')
class Integer(Descriptor):
    """
    This value must be an integer, or be converted to one
    """
    BASIC_MAKER = int


@register_custom_descriptor('float')
class Float(Descriptor):
    """
    This value must be a float, or be converted to one
    """
    BASIC_MAKER = float


@register_custom_descriptor('str')
class String(Descriptor):
    """
    This value must be a string, or be converted to one
    """
    BASIC_MAKER = str


class Regexp(String):
    """
    Base class for declaring regexp-based descriptors. Overload it's attribute REGEXP. Use as
    following:

    >>> class IPv6(Regexp):
    >>>     REGEXP = '(([0-9a-f]{1,4}:)' ...
    """
    __slots__ = ('regexp',)

    REGEXP = r'.*'

    def __init__(self):
        super().__init__()
        if isinstance(self.REGEXP, str):
            self.regexp = re.compile(self.REGEXP)
        else:
            self.regexp = self.REGEXP

    def __call__(self, value: ConfigDictValue) -> str:
        value = super(Regexp, self).__call__(value)

        match = self.regexp.match(value)
        if not match:
            raise ConfigurationValidationError('value does not match %s' % (self.REGEXP.pattern,),
                                               value)

        return match.group(0)

    def __str__(self):
        return 'Regexp(%s)' % (self.REGEXP.pattern,)


@register_custom_descriptor('ipv4')
class IPv4(Regexp):
    """
    This must be a valid IPv4 address (no hostnames allowed)
    """
    REGEXP = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'



from satella.coding.structures import Heap
from satella.coding.typing import K, V
import typing as tp


class merge_list(tp.Iterator[tp.Tuple[K, V]]):
    """
    Merge two sorted lists.

    This is an iterator which consumes elements as they are required.

    Each list must be of type tuple/2 with the first element being the key.
    The list has to be sorted by this value, ascending.

    When the algorithm encounters two identical keys, it calls merge_function on it's
    result and inserts the result.

    :param lists: lists to sort
    :param merge_function: a callable that accepts two pieces of the tuple and returns a result
    :return: an resulting iterator
    """
    __slots__ = ('merge_func', 'its', 'heap', 'available_lists',
                 'k', 'v', 'i', 'closed')

    def __iter__(self):
        return self

    def __init__(self, *lists: tp.Iterator[tp.Tuple[K, V]],
                 merge_function: tp.Callable[[V, V], V]):
        self.its = [iter(y) for y in lists]
        self.merge_func = merge_function
        self.available_lists = set()
        self.heap = Heap()
        self.closed = False

        for i, it in enumerate(self.its):
            try:
                self.heap.push((*next(it), i))
                self.available_lists.add(i)
            except StopIteration:
                pass
        try:
            self.k, self.v, self.i = self.pop()
        except IndexError:
            self.closed = True

    def __next__(self):
        if self.closed:
            raise StopIteration()

        try:
            k2, v2, i2 = self.pop()
        except IndexError:
            self.closed = True
            return self.k, self.v

        if k2 == self.k:
            self.v = self.merge_func(self.v, v2)
            return next(self)

        try:
            return self.k, self.v
        finally:
            self.k, self.v, self.i = k2, v2, i2

    def pop(self):
        k, v, i = self.heap.pop()
        if i in self.available_lists:
            try:
                k2, v2 = next(self.its[i])
                self.heap.push((k2, v2, i))
            except StopIteration:
                self.available_lists.remove(i)
        return k, v, i




import typing as tp

from satella.coding.structures import Proxy
from satella.coding.typing import V

__all__ = ['call_if_nnone', 'iterate_if_nnone', 'Optional']


def iterate_if_nnone(iterable: tp.Optional[tp.Iterable]) -> tp.Iterable:
    """
    Return a generator iterating over every element of iterable if it's not None.

    If it's None, return an empty generator

    :param iterable: iterable to iterate over
    :return: an empty generator if iterable is none, else an iterator over iterable
    """
    if iterable is not None:
        return iter(iterable)
    else:
        return iter(tuple())


def call_if_nnone(clbl: tp.Optional[tp.Callable[..., V]], *args, **kwargs) -> tp.Optional[V]:
    """
    Call clbl with provided arguments, but only if clbl is not None.

    If it's None, then None will be returned. Else, the result of this callable will be returned.

    :param clbl: callable to call, or a None
    :param args: positional arguments to provide to clbl
    :param kwargs: keyword arguments to provide to clbl
    :return: return value or None
    """
    if clbl is not None:
        return clbl(*args, **kwargs)
    else:
        return None


class Optional(Proxy):
    """
    A wrapper for your classes, that does nothing if the object passed in is None.
    It will return an empty Optional in that case.

    Usage example:

    >>> may_be_none = None
    >>> Optional(may_be_none).cancel().result()

    So far operations supported:

    * calling
    * getattr
    * getitem/setitem/delitem
    * testing for truth
    """
    __slots__ = ()

    def __init__(self, obj):
        super().__init__(obj)

    def __getattr__(self, item):
        if getattr(self, '_Proxy__obj') is None:
            return EMPTY_OPTIONAL
        else:
            return super().__getattr__(item)

    def __call__(self, *args, **kwargs):
        if getattr(self, '_Proxy__obj') is None:
            return EMPTY_OPTIONAL
        else:
            return super().__call__(*args, **kwargs)

    def __bool__(self):
        if getattr(self, '_Proxy__obj') is None:
            return False
        else:
            return super().__bool__()

    def __getitem__(self, item):
        if getattr(self, '_Proxy__obj') is None:
            return EMPTY_OPTIONAL
        else:
            return super().__getitem__(item)

    def __setitem__(self, key, value) -> None:
        if getattr(self, '_Proxy__obj') is None:
            return
        return super().__setitem__(key, value)

    def __delitem__(self, key) -> None:
        if getattr(self, '_Proxy__obj') is None:
            return
        return super().__delitem__(key)


EMPTY_OPTIONAL = Optional(None)



import binascii
import codecs
import typing as tb

from satella.coding.recast_exceptions import rethrow_as
from satella.exceptions import ConfigurationError
from .base import BaseSource

__all__ = [
    'FormatSource', 'FORMAT_SOURCES'
]

FORMAT_SOURCES = []  # sources capable of parsing a text format


def register_format_source(source):
    source_name = source.__name__
    __all__.append(source_name)
    FORMAT_SOURCES.append(source_name)

    return source


def _override_me(key):
    raise NotImplementedError('override me')


class FormatSource(BaseSource):
    __slots__ = ('root', 'encoding')

    TRANSFORM = _override_me
    BASE_EXCEPTIONS = [TypeError, UnicodeDecodeError, ValueError,
                       binascii.Error, LookupError]
    EXTRA_EXCEPTIONS = []

    def __init__(self, root: tb.Union[bytes, str], encoding: str = 'utf-8'):
        """
        :param root: content
        :type root: if bytes, will be decoded with given encoding'
        """
        super().__init__()
        self.root = root  # type: BaseSource
        self.encoding = encoding  # type: str

    def provide(self) -> dict:
        cls = self.__class__

        with rethrow_as(tuple(cls.BASE_EXCEPTIONS + cls.EXTRA_EXCEPTIONS),
                        ConfigurationError):
            if isinstance(self.root, bytes):
                self.root = codecs.decode(self.root, self.encoding)
                if isinstance(self.root, bytes):  # probably base64 encoded
                    self.root = self.root.decode('utf-8')

            ret_val = cls.TRANSFORM(self.root)
            if not isinstance(ret_val, dict):
                raise ConfigurationError(
                    'provider was unable to generate a text volume')
            else:
                return ret_val


try:
    import ujson


    @register_format_source
    class JSONSource(FormatSource):
        """
        Loads JSON strings
        """
        TRANSFORM = ujson.loads
        EXTRA_EXCEPTIONS = [ValueError]
except ImportError:
    import json


    @register_format_source
    class JSONSource(FormatSource):
        """
        Loads JSON strings
        """
        TRANSFORM = json.loads
        EXTRA_EXCEPTIONS = [json.JSONDecodeError]

try:
    import yaml
except ImportError:
    pass
else:
    @register_format_source
    class YAMLSource(FormatSource):
        """
        Loads YAML strings
        """
        EXTRA_EXCEPTIONS = [yaml.YAMLError]
        TRANSFORM = lambda data: yaml.load(data, Loader=yaml.Loader)

try:
    import toml
except ImportError:
    pass
else:
    @register_format_source
    class TOMLSource(FormatSource):
        """
        Loads TOML strings
        """
        EXTRA_EXCEPTIONS = [toml.TomlDecodeError]
        TRANSFORM = toml.loads



import ctypes
import platform
import threading
import time
import typing as tp
import warnings
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future
from threading import Condition as PythonCondition

from satella.coding.decorators import wraps
from satella.time import measure
from ..typing import ExceptionList
from ...exceptions import ResourceLocked, WouldWaitMore


def call_in_separate_thread(*t_args, **t_kwargs):
    """
    Decorator to mark given routine as callable in a separate thread.

    The decorated routine will return a Future that is waitable to get the result
    (or the exception) of the function.

    The arguments given here will be passed to thread's constructor, so use like:

    >>> @call_in_separate_thread(daemon=True)
    >>> def handle_messages():
    >>>     while True:
    >>>         ...
    """

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs) -> Future:
            class MyThread(threading.Thread):
                def __init__(self):
                    self.future = Future()
                    super().__init__(*t_args, **t_kwargs)

                def run(self):
                    if not self.future.set_running_or_notify_cancel():
                        return
                    try:
                        res = fun(*args, **kwargs)
                        self.future.set_result(res)
                    except Exception as e:
                        self.future.set_exception(e)

            t = MyThread()
            t.start()
            return t.future

        return inner

    return outer


class Condition(PythonCondition):
    """
    A wrapper to faciliate easier usage of Pythons' threading.Condition.

    There's no need to acquire the underlying lock, as wait/notify/notify_all do it for you.

    This happens to sorta not work on PyPy. Use at your own peril. You have been warned.
    """

    def notifyAll(self) -> None:
        """
        Deprecated alias for notify_all

        .. deprecated:: 2.14.22
        """
        warnings.warn('Use notify_all instead', DeprecationWarning)
        self.notify_all()

    def wait(self, timeout: tp.Optional[float] = None,
             dont_raise: bool = False) -> None:
        """
        Wait for condition to become true.

        :param timeout: timeout to wait. None is default and means infinity
        :param dont_raise: if True, then WouldWaitMore won't be raised
        :raises ResourceLocked: unable to acquire the underlying lock within specified timeout.
        :raises WouldWaitMore: wait's timeout has expired
        """
        if timeout is not None:
            if timeout < 0:
                timeout = 0

        with measure(timeout=timeout) as measurement:
            if timeout is None:
                self.acquire()
            else:
                if not self.acquire(timeout=measurement.time_remaining):
                    raise ResourceLocked('internal lock locked')

            try:
                if timeout is None:
                    super().wait()
                else:
                    if not super().wait(timeout=measurement.time_remaining):
                        if not dont_raise:
                            raise WouldWaitMore('wait was not notified')
            finally:
                self.release()

    def notify_all(self) -> None:
        """
        Notify all threads waiting on this Condition
        """
        with self._lock:
            super().notify_all()

    def notify(self, n: int = 1) -> None:
        """
        Notify n threads waiting on this Condition

        :param n: amount of threads to notify
        """
        with self._lock:
            super().notify(n=n)


class SingleStartThread(threading.Thread):
    """
    A thread that keeps track of whether it's .start() method was called, and does nothing
    if it's called second or so time.
    """

    def __init__(self, *args, **kwargs):
        self.__started = False
        super().__init__(*args, **kwargs)

    def start(self) -> 'SingleStartThread':
        """
        No-op when called second or so time. The first time it starts the thread.

        :return: self
        """
        if self.__started:
            return
        self.__started = True
        super().start()
        return self


class BogusTerminableThread:
    """
    A mock object that implements threading interface but does nothing
    """

    def __init__(self):
        self.running = False
        self.terminated = False
        self.daemon = True

    def is_alive(self):
        return not self.terminated and self.running

    def start(self):
        """
        :raises RuntimeError: thread already terminated or already running
        """
        if self.terminated:
            raise RuntimeError('Thread already terminated')
        if self.running:
            raise RuntimeError('Thread already running')
        self.running = True

    def terminate(self):
        self.terminated = True

    def join(self, timeout=None):
        if self.terminated:
            self.running = False


class TerminableThread(threading.Thread):
    """
    Class that will execute something in a loop unless terminated. Use like:

    >>> class MeGrimlock(TerminableThread):
    >>>     def loop(self):
    >>>         ... do your operations ..
    >>> a = MeGrimlock().start()
    >>> a.terminate().join()

    Flag whether to terminate is stored in **self.terminating**.

    If you decide to override run(), you got to check periodically for **self._terminating**
    to become true. If it's true, then a termination request was received, and the thread should
    terminate itself.
    If you decide to use the loop/cleanup interface, you don't need to do so, because it will
    be automatically checked for you before each loop() call.

    You may also use it as a context manager. Entering the context will start the thread, and
    exiting it will .terminate().join() it, in the following way:

    >>> a = MeGrimlock()
    >>> with a:
    >>>     ...
    >>> self.assertFalse(a.is_alive())
    """

    def __init__(self, *args, terminate_on: tp.Optional[ExceptionList] = None,
                 **kwargs):
        """
        Note that this is called in the constructor's thread. Use .prepare() to
        run statements that should be ran in new thread.

        :param terminate_on: if provided, and
            :meth:`~satella.coding.concurrent.TerminableThread.loop` throws one of it,
            swallow it and terminate the thread by calling
            :meth:`~satella.coding.concurrent.TerminableThread.terminate`. Note that the
            subclass check will be done via `isinstance` so you can use the metaclass magic :)
        """
        super().__init__(*args, **kwargs)
        self._terminating = False  # type: bool
        self._terminate_on = terminate_on

    @property
    def terminating(self) -> bool:
        """Return whether a termination of this thread was requested"""
        return self._terminating

    def prepare(self) -> None:
        """
        This is called before the .loop() looping loop is entered.

        This is invoked already in a separate thread.
        """

    def loop(self) -> None:
        """
        Run one iteration of the loop. Meant to be overrided. You do not need to override it
        if you decide to override run() through.

        This should block for as long as a single check will take, as termination checks take place
        between calls.

        Note that if it throws one of the exceptions given in `terminate_on` this thread will
        terminate cleanly, whereas if it throws something else, the thread will be terminated with
        a traceback.
        """

    def start(self) -> 'TerminableThread':
        """
        Start the execution of this thread
        :return: this thread
        """
        super().start()
        return self

    def run(self) -> None:
        """
        Calls self.loop() indefinitely, until terminating condition is met
        """
        try:
            self.prepare()
            while not self._terminating:
                try:
                    self.loop()
                except Exception as e:
                    if self._terminate_on is not None:
                        if isinstance(e, self._terminate_on):
                            self.terminate()
                    else:
                        raise
        except SystemExit:
            pass
        finally:
            self.cleanup()

    def cleanup(self):
        """
        Called after thread non-forced termination, in the thread's context.

        The default implementation does nothing.
        """

    def __enter__(self):
        """Returns self"""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate().join()
        return False

    def safe_sleep(self, interval: float, wake_up_each: float = 2) -> None:
        """
        Sleep for interval, waking up each wake_up_each seconds to check if terminating,
        finish earlier if is terminating.

        This will do *the right thing* when passed a negative interval.

        :param interval: Time to sleep in total
        :param wake_up_each: Amount of seconds to wake up each
        :raises SystemExit: thread is terminating
        """
        t = 0
        while t < interval and not self._terminating:
            remaining_to_sleep = min(interval - t, wake_up_each)
            time.sleep(remaining_to_sleep)
            t += remaining_to_sleep
        if self._terminating:
            raise SystemExit()

    def terminate(self, force: bool = False) -> 'TerminableThread':
        """
        Signal this thread to terminate.

        Forcing, if requested, will be done by injecting a SystemExit exception into target
        thread, so the thread must acquire GIL. For example, following would not be interruptable:

        >>> time.sleep(1000000)

        Note that calling force=True on PyPy won't work, and NotImplementedError will be raised
        instead.

        :param force: Whether to force a quit
        :return: self
        :raises RuntimeError: when something goes wrong with the underlying Python machinery
        :raises NotImplementedError: force=True was used on PyPy
        """
        self._terminating = True
        if force:
            if platform.python_implementation() == 'PyPy':
                raise NotImplementedError('force=True was made on PyPy')
            ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident),
                                                             ctypes.py_object(SystemExit))
            if ret == 0:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Zero threads killed!')
            elif ret > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Multiple threads killed!')

        return self


class IntervalTerminableThread(TerminableThread, metaclass=ABCMeta):
    """
    A TerminableThread that calls .loop() once per x seconds.

    If executing .loop() takes more than x seconds, on_overrun() will be called.

    :param seconds: time that a single looping through should take. This will
        include the time spent on calling .loop(), the rest of this time will
        be spent safe_sleep()ing.
    """

    def __init__(self, seconds: float, *args, **kwargs):
        self.seconds = seconds
        super().__init__(*args, **kwargs)

    @abstractmethod
    def loop(self) -> None:
        """
        Override me!
        """

    def on_overrun(self, time_taken: float) -> None:
        """
        Called when executing .loop() takes more than x seconds.

        Called each cycle.

        :param time_taken: how long did calling .loop() take
        """

    def run(self):
        self.prepare()
        while not self._terminating:
            with measure() as measurement:
                self.loop()
            time_taken = measurement()
            time_to_sleep = self.seconds - time_taken
            if time_to_sleep < 0:
                self.on_overrun(time_taken)
            else:
                self.safe_sleep(time_to_sleep)
        self.cleanup()



import time
import typing as tp
from concurrent.futures import Executor, ThreadPoolExecutor, ProcessPoolExecutor

from satella.coding.concurrent.sync import sync_threadpool
from satella.coding.concurrent.monitor import Monitor
from satella.coding.recast_exceptions import silence_excs
from satella.coding.typing import V, K


class ExclusiveWritebackCache(tp.Generic[K, V]):
    """
    A dictionary implementing an exclusive write-back cache. By exclusive it is understood
    that only this object will be modifying the storage.

    :param write_method: a blocking callable (key, value) that writes the value to underlying
        storage.
    :param read_method: a blocking callable (key) -> value that retrieves the piece of data from
        the cache, or throws a KeyError to signal that the value does not exist
    :param delete_method: optional, a blocking callable (key) that erases the data from the storage.
        If not given, it will be a TypeError to delete the data from this storage
    :param executor: an executor to execute the calls with. If None (default) is given, a
        ThreadPoolExecutor with 4 workers will be created
    :param no_concurrent_executors: number of concurrent jobs that the executor is able
        to handle. This is used by sync()
    :param store_key_errors: whether to remember KeyErrors raised by read_method
    """
    __slots__ = ('executor', 'read_method', 'write_method', 'delete_method',
                 'no_concurrent_executors', 'in_cache', 'cache_lock',
                 'cache', 'operations', 'store_key_errors')

    def __init__(self, write_method: tp.Callable[[K, V], None],
                 read_method: tp.Callable[[K], V],
                 delete_method: tp.Optional[tp.Callable[[K], None]] = None,
                 executor: tp.Optional[Executor] = None,
                 no_concurrent_executors: tp.Optional[int] = None,
                 store_key_errors: bool = True
                 ):
        if executor is None:
            self.executor = ThreadPoolExecutor(max_workers=4)
        else:
            self.executor = executor
        self.store_key_errors = store_key_errors
        self.write_method = write_method
        self.delete_method = delete_method
        self.read_method = read_method
        self.no_concurrent_executors = no_concurrent_executors or 4
        self.in_cache = set()
        self.cache_lock = Monitor()
        self.cache = {}
        self.operations = 0

    def get_queue_length(self) -> int:
        """
        Return current amount of entries waiting for writeback
        """
        # noinspection PyProtectedMember
        if isinstance(self.executor, ThreadPoolExecutor):
            return self.executor._work_queue.qsize()
        elif isinstance(self.executor, ProcessPoolExecutor):
            return self.executor._call_queue.qsize()
        else:
            return 0

    def sync(self, timeout: tp.Optional[float] = None) -> None:
        """
        Wait until current tasks are complete.

        :param timeout: timeout to wait. None means wait indefinitely.
        :raises WouldWaitMore: if timeout has expired
        """
        while self.get_queue_length() > 0:
            time.sleep(0.1)

        def fix():
            return

        sync_threadpool(self.executor, max_wait=timeout)

    def _operate(self):
        self.operations += 1
        if self.operations > 100:
            with self.cache_lock:
                self.in_cache = set(self.cache.keys())
                self.operations = 0

    def __getitem__(self, item: K) -> V:
        self._operate()
        if item not in self.in_cache:
            try:
                value = self.executor.submit(self.read_method, item).result()
                with self.cache_lock:
                    self.in_cache.add(item)
                    self.cache[item] = value
                return value
            except KeyError:
                if self.store_key_errors:
                    with self.cache_lock:
                        self.in_cache.add(item)
                raise
        else:
            if item not in self.cache and self.store_key_errors:
                raise KeyError()
            else:
                return self.cache[item]

    def __delitem__(self, key: K) -> None:
        if self.delete_method is None:
            raise TypeError('Cannot delete from this writeback cache!')
        if self.store_key_errors:
            with self.cache_lock:
                self.in_cache.add(key)
        with silence_excs(KeyError):
            del self.cache[key]
        self.executor.submit(self.delete_method, key)
        self._operate()

    def __setitem__(self, key: K, value: V) -> None:
        with self.cache_lock:
            self.cache[key] = value
            self.in_cache.add(key)
        self.executor.submit(self.write_method, key, value)
        self._operate()



import enum
import typing as tp
import warnings
from inspect import Parameter, signature
from queue import Queue

from satella.coding.recast_exceptions import rethrow_as


def enum_value(value):
    """
    If value is an enum, extract and return it's value.

    Otherwise, return it as-is.

    :param value: value to extract enum from
    :return: value
    """
    if isinstance(value, enum.Enum):
        return value.value
    return value


def contains(needle, haystack) -> bool:
    """
    A syntactic sugar for the following:

    >>> for item in haystack:
    >>>     if needle == item:
    >>>         return True
    >>> return False

    Note that this is very like Python's in operator, however it's not quite same, since
    in doesn't involve the __eq__ operator at every step!

    This function for example allows you to circumvent Python's limitations concerning
    :class:`~satella.coding.structures.ComparableEnum`

    :param needle: needle to check for
    :param haystack: haystack to check against
    :return: whether haystack contains the element
    """
    for item in haystack:
        if needle == item:
            return True
    return False


class Closeable:
    """
    A class that needs to clean up it's own resources.

    It's destructor calls .close(). Use like this:

    >>> class MyClose(Closeable):
    >>>     def close(self):
    >>>         if super().close():
    >>>             .. clean up ..

    Can be also used as a context manager, with close() called upon __exit__.

    You should extend both __init__ and close()
    """
    __slots__ = ('__finalized', )

    def __init__(self):
        self.__finalized = False

    def close(self) -> bool:
        """
        Check if the resource needs cleanup, and clean up this resource.

        Use like this:

        >>> class MyClose(Closeable):
        >>>     def close(self):
        >>>         if super().close():
        >>>             .. clean up ..

        :return: whether the cleanup should proceed
        :raises RuntimeError: the constructor was not invoked
        """
        try:
            return not self.__finalized
        except AttributeError:
            raise RuntimeError('__finalized not found, did you forget to call the constructor?')
        finally:
            self.__finalized = True

    def __del__(self) -> None:
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def queue_iterator(queue: Queue) -> tp.Iterator:
    """
    Syntactic sugar for

    >>> while queue.qsize() > 0:
    >>>     yield queue.get()
    """
    while queue.qsize() > 0:
        yield queue.get()


def update_if_not_none(dictionary: tp.Dict, key: tp.Hashable, value) -> tp.Dict:
    """
    Deprecated alias for :func:`update_key_if_none`

    .. deprecated:: 2.14.22
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0,'
                  'use update_key_if_not_none instead', DeprecationWarning)
    return update_key_if_none(dictionary, key, value)


def chain_callables(callable1: tp.Callable, callable2: tp.Callable) -> tp.Callable:
    """
    Link two callables together. callable2, if it takes an argument, will receive
    callables'1 result, and if it takes no arguments it will received nothing.

    :param callable1: first callable to call
    :param callable2: callable to call with callable1's result
    :return: result of callable2
    """
    def inner(*args, **kwargs):
        res = callable1(*args, **kwargs)
        try:
            res = callable2(res)
        except TypeError as e:
            if 'positional arguments but' in e.args[0] \
                    and 'was given' in e.args[0] and 'takes' in e.args[0]:
                res = callable2()
            else:
                raise
        return res
    return inner


def source_to_function(src: tp.Union[tp.Callable, str]) -> tp.Callable[[tp.Any], tp.Any]:
    """
    If src is callable, return it as-is
    Transform a string containing a Python expression with a variable x to a lambda.

    It will be treated as if it was appended to 'lambda x: '

    WARNING: Do not run untrusted data. Familiarize yourself with the dangers of passing
    unvalidated data to exec() or eval()!

    :param src: a callable or a Python string expression
    :return: a callable
    """
    if isinstance(src, str):
        q = dict(globals())
        exec('_precond = lambda x: ' + src, q)
        return q['_precond']
    else:
        return src


def update_attr_if_none(obj: object, attr: str, value: tp.Any,
                        on_attribute_error: bool = True,
                        if_value_is_not_none: bool = False) -> object:
    """
    Updates the object attribute, if it's value is None, or if
    it yields AttributeError (customizable as per on_attribute_error parameter)

    :param obj: object to alter
    :param attr: attribute to set
    :param value: value to set
    :param on_attribute_error: whether to proceed with setting the value on
        AttributeError while trying to read given attribute. If False, AttributeError
        will be raised.
    :param if_value_is_not_none: update object unconditionally, if only value is not None
    :return: obj
    """
    if if_value_is_not_none:
        if value is not None:
            setattr(obj, attr, value)
    else:
        try:
            val = getattr(obj, attr)
            if val is None:
                setattr(obj, attr, value)
        except AttributeError:
            if on_attribute_error:
                setattr(obj, attr, value)
            else:
                raise
    return obj


class _BLANK:
    pass


_BlankType = tp.Type[_BLANK]


def update_key_if_true(dictionary: tp.Dict, key: tp.Hashable, value: tp.Any,
                       flag: tp.Union[bool, _BlankType] = _BLANK) -> tp.Dict:
    """
    If flag is True, execute dictionary[key] = value

    :param dictionary: dictionary to mutate
    :param key: dictionary key to use
    :param value: dictionary value to set
    :param flag: whether to execute the setting operation. If let at default,
        flag will be calculated from boolean of the value
    :return: the dict itself
    """
    if flag is _BLANK:
        flag = bool(value)
    if flag:
        dictionary[key] = value
    return dictionary


def get_arguments(function: tp.Callable, *args, **kwargs) -> \
        tp.Dict[str, tp.Any]:
    """
    Return local variables that would be defined for given function if called with
    provided arguments.

    Note that this function will not return the "self" argument of methods
    and it won't return the class of "cls" of classmethods.

    :param function: callable to examine
    :param args: arguments to provide
    :param kwargs: keyword arguments to provide
    :return: a dictionary of local variables with their values, as they would
        appear in function if called with provided arguments
    :raise TypeError: the dictionary cannot be created with provided arguments
    """
    return _get_arguments(function, False, *args, **kwargs)


@rethrow_as(IndexError, TypeError)
def _get_arguments(function: tp.Callable, special_behaviour: bool, *args, **kwargs):
    """
    :param special_behaviour: enable special behaviour to be used in for_arguments
    """
    sig = signature(function)
    params = sig.parameters.values()
    local_vars = {}

    positionals = [param for param in reversed(params) if
                   param.kind in (Parameter.POSITIONAL_OR_KEYWORD,
                                  Parameter.POSITIONAL_ONLY,
                                  Parameter.VAR_POSITIONAL)]
    args = list(reversed(args))

    arguments_left = set(param.name for param in params)
    while len(positionals):
        arg = positionals.pop()
        arg_kind = arg.kind
        arg_name = arg.name
        if arg_kind == Parameter.VAR_POSITIONAL and not special_behaviour:
            local_vars[arg_name] = tuple(reversed(args))
        else:
            try:
                v = args.pop()
                arguments_left.remove(arg_name)
            except IndexError:
                try:
                    if arg.default != Parameter.empty:
                        raise AttributeError()
                    else:
                        break
                except (AttributeError, TypeError):
                    v = arg.default
            local_vars[arg_name] = v

    keywords = [param for param in params if param.kind in (Parameter.POSITIONAL_OR_KEYWORD,
                                                            Parameter.KEYWORD_ONLY,
                                                            Parameter.VAR_KEYWORD)
                and param.name in arguments_left]

    for keyword in keywords:
        keyword_name = keyword.name
        if keyword.kind == Parameter.VAR_KEYWORD and not special_behaviour:
            local_vars[keyword_name] = kwargs
        else:
            try:
                v = kwargs.pop(keyword_name)
            except KeyError:
                try:
                    if Parameter.empty == keyword.default:
                        if special_behaviour:
                            v = None
                        else:
                            raise TypeError('Not enough keyword arguments')
                    else:
                        v = keyword.default
                except (AttributeError, TypeError):
                    continue    # comparison was impossible

            local_vars[keyword_name] = v

    for param in params:
        param_name = param.name
        if param_name not in local_vars:
            if special_behaviour:
                local_vars[param_name] = None
            else:
                raise TypeError('Not enough keyword arguments')

    return local_vars


def call_with_arguments(function: tp.Callable, arguments: tp.Dict[str, tp.Any]) -> tp.Any:
    """
    Call a function, but with giving it arguments via a dictionary.

    Dictionary should be a mapping of argument name to it's value.

    :param function: function to call
    :param arguments:
        a dict of arguments : argument name => argument value.
        This dictionary will be modified!
    :return: return value of the function
    :raise TypeError: too few arguments, or some arguments required were missing
    :raise ValueError: too many arguments given
    """
    args = []
    kwargs = {}
    for param in signature(function).parameters.values():
        param_name = param.name
        param_kind = param.kind
        if param_name not in arguments:
            if param_kind in (Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL):
                continue
            elif param.default == Parameter.empty:
                raise TypeError('Argument %s not found' % (param_name,))
            else:
                continue

        if param_kind == Parameter.POSITIONAL_ONLY or param_kind == Parameter.POSITIONAL_OR_KEYWORD:
            args.append(arguments.pop(param_name))
        elif param_kind == Parameter.VAR_POSITIONAL:
            args.extend(arguments.pop(param_name))
        elif param_kind == Parameter.KEYWORD_ONLY:
            kwargs[param_name] = arguments.pop(param_name)
        elif param_kind == Parameter.VAR_KEYWORD:
            kwargs.update(arguments.pop(param_name))
        else:
            raise TypeError('Unknown parameter type')

    if arguments:
        raise ValueError('Too many arguments provided')

    return function(*args, **kwargs)


def update_key_if_none(dictionary: tp.Dict, key: tp.Hashable, value) -> tp.Dict:
    """
    This is deprecated. Please use update_key_if_not_none instead!

    .. deprecated:: 2.14.22
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0, use '
                  'update_key_if_not_none instead', DeprecationWarning)
    return update_key_if_not_none(dictionary, key, value)


def update_key_if_not_none(dictionary: tp.Dict, key: tp.Union[tp.Hashable, tp.Dict],
                           value: tp.Union[tp.Any, _BlankType] = _BLANK) -> tp.Dict:
    """
    Syntactic sugar for

    >>> if value is not None:
    >>>     dictionary[key] = value

    If value is passed, else

    >>> for key, value in key.items():
    >>>     if value is not None:
    >>>         dictionary[key] = value

    :param dictionary: dictionary to update
    :param key: key to use or a dictionary of items
    :param value: value to use
    :return: the dictionary itself
    """
    if value is _BLANK:
        for key, val in key.items():
            if val is not None:
                dictionary[key] = val
    elif value is not None:
        dictionary[key] = value
    return dictionary



import inspect
import typing as tp
from inspect import Parameter


def extract_type_signature_from(fun: tp.Callable) -> tp.Tuple[type, ...]:
    sign = []
    params = inspect.signature(fun).parameters
    for parameter in params.values():
        if parameter.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
            if parameter.annotation == Parameter.empty:
                sign.append(None)
            else:
                sign.append(parameter.annotation)
    return tuple(sign)


# Taken from https://stackoverflow.com/questions/28237955/same-name-for-classmethod-and-
# instancemethod
class class_or_instancemethod(classmethod):
    """
    A decorator to make your methods both classmethods (they will receive an instance of type
    as their first argument) or normal methods (they will receive an instance of their type).

    Use like:

    >>> class MyClass:
    >>>     @class_or_instancemethod
    >>>     def method(self_or_cls):
    >>>         if isinstance(self_or_cls, MyClass):
    >>>             # method code
    >>>         else:
    >>>             # classmethod code
    """

    def __get__(self, instance, type_):
        descr_get = super().__get__ if instance is None else self.__func__.__get__
        return descr_get(instance, type_)


class overload:
    """
    A class used for method overloading.

    Note that methods can be only overloaded by their positional, or positional-and-keyword
    arguments. Overload distinguishment will be done at the level of positional arguments only.

    Note that typing checks will be done via isinstance().

    Use like this:

    >>> @overload
    >>> def what_type(x: str):
    >>>     print('String')
    >>> @what_type.overload
    >>> def what_type(x: int):
    >>>     print('Int')
    """

    def __init__(self, fun: tp.Callable):
        self.type_signatures_to_functions = {
            extract_type_signature_from(fun): fun
        }  # type: tp.Dict[tp.Tuple[type, ...], tp.Callable]
        if hasattr(fun, '__doc__'):
            self.__doc__ = fun.__doc__

    def overload(self, fun):
        """
        :raises ValueError: this signature already has an overload
        """
        sign = extract_type_signature_from(fun)
        if sign in self.type_signatures_to_functions:
            f = self.type_signatures_to_functions[sign]
            raise ValueError('Method of this signature is already overloaded with %s' % (f,))
        self.type_signatures_to_functions[sign] = fun
        return self

    def __call__(self, *args, **kwargs):
        """
        Call one of the overloaded functions.

        :raises TypeError: no type signature matched
        """
        for sign, fun in self.type_signatures_to_functions.items():
            try:
                for type_, arg in zip(sign, args):
                    if type_ is None:
                        continue
                    if not isinstance(arg, type_):
                        raise ValueError()

                return fun(*args, **kwargs)
            except ValueError:
                pass
        raise TypeError('No matching functions found')



import typing as tp

from .monitor import Monitor
from .thread import Condition
from ...exceptions import WouldWaitMore
from ...time import measure

Number = tp.Union[int, float]


class AtomicNumber(Monitor):
    """
    An atomic number. Note that the class is not hashable and for a reason, since it's value
    might change in time. So in this case this is more of like a container for numbers.

    Treat it like a normal number, except all operations are executed atomically.

    You can also wait for it to change it's value, via wait().

    You change it's value in the following way:

    >>> a = AtomicNumber()
    >>> a += 2

    Note that if the number if used in an expression, such as

    >>> b = a + 2

    Then a normal number will be returned
    """
    __slots__ = ('value', 'condition')

    def __init__(self, v: Number = 0):
        super().__init__()
        self.value = v
        self.condition = Condition()

    def wait(self, timeout: tp.Optional[float] = None, throw_exception: bool = True):
        """
        Block until the atomic number changes it's value.

        :param timeout: maximum time to wait. None means wait indefinitely
        :param throw_exception: whether to throw WouldWaitMore on timeout
        :raises WouldWaitMore: the value hasn't changed within the timeout
        """
        try:
            self.condition.wait(timeout)
        except WouldWaitMore:
            if throw_exception:
                raise

    @Monitor.synchronized
    def __repr__(self) -> str:
        return 'AtomicNumber(%s)' % (self.value,)

    @Monitor.synchronized
    def __and__(self, other: Number) -> Number:
        return self.value & other

    @Monitor.synchronized
    def __or__(self, other: Number) -> Number:
        return self.value | other

    @Monitor.synchronized
    def __xor__(self, other: Number) -> Number:
        return self.value ^ other

    @Monitor.synchronized
    def __rand__(self, other: Number) -> Number:
        return other & self.value

    @Monitor.synchronized
    def __ror__(self, other: Number) -> Number:
        return other | self.value

    @Monitor.synchronized
    def __rxor__(self, other: Number) -> Number:
        return other ^ self.value

    @Monitor.synchronized
    def __iand__(self, other: Number) -> 'AtomicNumber':
        self.value &= other
        return self

    @Monitor.synchronized
    def __ior__(self, other: Number) -> 'AtomicNumber':
        self.value |= other
        return self

    @Monitor.synchronized
    def __ixor__(self, other: Number) -> 'AtomicNumber':
        self.value ^= other
        return self

    @Monitor.synchronized
    def __str__(self) -> str:
        return str(self.value)

    @Monitor.synchronized
    def __gt__(self, other: Number):
        return self.value > other

    @Monitor.synchronized
    def __lshift__(self, other: int) -> Number:
        return self.value << other

    @Monitor.synchronized
    def __rshift__(self, other: int) -> Number:
        return self.value >> other

    @Monitor.synchronized
    def __rlshift__(self, other: int) -> Number:
        return other << self.value

    @Monitor.synchronized
    def __rrshift__(self, other: int) -> Number:
        return other >> self.value

    @Monitor.synchronized
    def __ilshift__(self, other: int) -> 'AtomicNumber':
        self.value <<= other
        return self

    @Monitor.synchronized
    def __irshift__(self, other: int) -> 'AtomicNumber':
        self.value >>= other
        return self

    @Monitor.synchronized
    def __ge__(self, other: Number):
        return self.value >= other

    @Monitor.synchronized
    def __lt__(self, other: Number):
        return self.value < other

    @Monitor.synchronized
    def __le__(self, other: Number):
        return self.value <= other

    @Monitor.synchronized
    def __iadd__(self, other: Number) -> 'AtomicNumber':
        self.value += other
        self.condition.notify_all()
        return self

    @Monitor.synchronized
    def __pow__(self, power: Number, modulo=None) -> Number:
        return pow(self.value, power, modulo)

    @Monitor.synchronized
    def __ipow__(self, other) -> 'AtomicNumber':
        self.value **= other
        return self

    @Monitor.synchronized
    def __radd__(self, other: Number) -> Number:
        return other + self.value

    @Monitor.synchronized
    def __rsub__(self, other: Number) -> Number:
        return other - self.value

    @Monitor.synchronized
    def __rmul__(self, other: Number) -> Number:
        return other * self.value

    @Monitor.synchronized
    def __rtruediv__(self, other: Number) -> Number:
        return other / self.value

    @Monitor.synchronized
    def __rfloordiv__(self, other: Number) -> Number:
        return other // self.value

    @Monitor.synchronized
    def __rpow__(self, other: Number) -> Number:
        return other ** self.value

    @Monitor.synchronized
    def __radd__(self, other: Number) -> Number:
        return other + self.value

    @Monitor.synchronized
    def __add__(self, other: Number) -> Number:
        return self.value + other

    @Monitor.synchronized
    def __sub__(self, other: Number) -> Number:
        return self.value - other

    @Monitor.synchronized
    def __mul__(self, other: Number) -> Number:
        return self.value * other

    @Monitor.synchronized
    def __truediv__(self, other: Number) -> Number:
        return self.value / other

    @Monitor.synchronized
    def __floordiv__(self, other: Number) -> Number:
        return self.value // other

    @Monitor.synchronized
    def __itruediv__(self, other: Number) -> 'AtomicNumber':
        self.value /= other
        self.condition.notify_all()
        return self

    @Monitor.synchronized
    def __ifloordiv__(self, other: Number) -> 'AtomicNumber':
        self.value //= other
        self.condition.notify_all()
        return self

    @Monitor.synchronized
    def __isub__(self, other: int) -> 'AtomicNumber':
        self.value -= other
        self.condition.notify_all()
        return self

    @Monitor.synchronized
    def __imul__(self, other: int) -> 'AtomicNumber':
        self.value *= other
        self.condition.notify_all()
        return self

    @Monitor.synchronized
    def __eq__(self, other: tp.Union['AtomicNumber', Number]) -> bool:
        if isinstance(other, AtomicNumber):
            with Monitor.acquire(other):
                return self.value == other.value
        else:
            return self.value == other

    @Monitor.synchronized
    def __int__(self) -> int:
        return int(self.value)

    @Monitor.synchronized
    def __float__(self) -> float:
        return float(self.value)

    @Monitor.synchronized
    def __bool__(self) -> bool:
        return bool(self.value)

    @Monitor.synchronized
    def __abs__(self) -> Number:
        return abs(self.value)

    @Monitor.synchronized
    def __neg__(self) -> Number:
        return -self.value

    @Monitor.synchronized
    def __pos__(self) -> Number:
        return +self.value

    def wait_until_equal(self, v: Number, timeout: tp.Optional[float] = None) -> None:
        """
        Wait until the value of this number equals v.

        :param v: value to compare this number against
        :param timeout: maximum time to wait
        :raise WouldWaitMore: timeout expired without the value becoming equal to target
        """
        if timeout is None:
            while True:
                if self == v:
                    break
                self.wait()
        else:
            with measure(timeout=timeout) as measurement:
                while not measurement.timeouted:
                    if self == v:
                        break
                    self.wait(measurement.time_remaining, throw_exception=False)

                with Monitor.acquire(self):
                    if self.value != v:
                        raise WouldWaitMore()



import itertools
import math
import typing as tp

from .base import EmbeddedSubmetrics, MetricLevel
from .measurable_mixin import MeasurableMixin
from .registry import register_metric
from ..data import MetricData, MetricDataCollection


@register_metric
class HistogramMetric(EmbeddedSubmetrics, MeasurableMixin):
    """
    A histogram, by  `Prometheus' <https://github.com/prometheus/client_python#histogram/>`_
    interpretation.
    
    :param buckets: buckets to add. First bucket will be from zero to first value, second from first
        value to second, last bucket will be from last value to infinity. So there are 
        len(buckets)+1 buckets. Buckets are expected to be passed in sorted!
    :param aggregate_children: whether to accept child calls to be later presented as total
    """
    __slots__ = ('bucket_limits', 'buckets', 'aggregate_children', 'count', 'sum')

    CLASS_NAME = 'histogram'

    def __init__(self, name: str, root_metric: 'Metric' = None,
                 metric_level: tp.Optional[MetricLevel] = None,
                 internal: bool = False,
                 buckets: tp.Sequence[float] = (.005, .01, .025, .05, .075, .1, .25, .5,
                                                .75, 1.0, 2.5, 5.0, 7.5, 10.0),
                 aggregate_children: bool = True, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, internal=internal, buckets=buckets,
                         aggregate_children=aggregate_children, *args, **kwargs)
        self.bucket_limits = list(buckets)  # type: tp.List[float]
        self.buckets = [0] * (len(buckets) + 1)  # type: tp.List[int]
        self.aggregate_children = aggregate_children  # type: bool
        self.count = 0  # type: int
        self.sum = 0.0  # type: float

    def _handle(self, value, **labels):
        self.count += 1
        self.sum += value

        if self.embedded_submetrics_enabled or labels:
            super()._handle(value, **labels)

            if not self.aggregate_children:
                return

        lower_bound = 0.0
        for index, upper_bound in itertools.chain(enumerate(self.bucket_limits),
                                                  [(len(self.bucket_limits), math.inf)]):
            if lower_bound <= value < upper_bound:
                self.buckets[index] += 1
            lower_bound = upper_bound

    def to_metric_data(self) -> MetricDataCollection:
        if self.embedded_submetrics_enabled:
            k = super().to_metric_data()
            if self.aggregate_children:
                mdc = self.containers_to_metric_data()
                mdc.postfix_with('total')
                mdc += MetricData(self.name + '.total.sum', self.sum, {}, self.get_timestamp())
                mdc += MetricData(self.name + '.total.count', self.count, {}, self.get_timestamp())
                k += mdc
            return k

        mdc = self.containers_to_metric_data()
        mdc += MetricData(self.name + '.sum', self.sum, self.labels, self.get_timestamp())
        mdc += MetricData(self.name + '.count', self.count, self.labels, self.get_timestamp())
        return mdc

    def containers_to_metric_data(self) -> MetricDataCollection:
        output = []
        lower_bound = 0.0
        for amount, upper_bound in zip(self.buckets,
                                       self.bucket_limits + [math.inf]):
            labels = self.labels.copy()
            labels.update(le=upper_bound,
                          ge=lower_bound)
            output.append(
                MetricData(self.name, amount, labels, self.get_timestamp(), self.internal))
            lower_bound = upper_bound
        return MetricDataCollection(output)



from .daemon import daemonize
from .misc import suicide, is_running_as_root
from .pidlock import PIDFileLock
from .signals import hang_until_sig

__all__ = [
    'daemonize',
    'PIDFileLock', 'hang_until_sig',
    'suicide', 'is_running_as_root'
]



import typing as tp
from collections import OrderedDict

T = tp.TypeVar('T', covariant=tp.Hashable)


class LRU(tp.Generic[T]):
    """
    A class to track least recently used objects.
    """

    def __init__(self):
        self.od = OrderedDict()

    def __contains__(self, item: T) -> bool:
        return item in self.od

    def __len__(self):
        return len(self.od)

    def add(self, item: T) -> None:
        self.od[item] = True

    def remove(self, item: T) -> None:
        del self.od[item]

    def mark_as_used(self, item: T):
        if item not in self.od:
            self.od[item] = True
        else:
            self.od.move_to_end(item)

    def get_item_to_evict(self) -> T:
        """
        Return a least recently used object
        """
        return self.od.popitem(False)[0]
